#ifndef CUDD_H
#define CUDD_H

#include "mublas_devel.h"

/* ***********************************************
   Error Free Translations
   *********************************************** */   

#define SPLITTER 134217729.0 //2^27+1

__device__ __forceinline__ void
cuTwoSum
	(double a, double b, double &s, double &e)
{
	double v;
	s = __dadd_rn (a, b);
	v = __dsub_rn (s, a);
	e = __dadd_rn ( __dsub_rn(a, __dsub_rn (s, v)), __dsub_rn (b, v));
}

__device__ __forceinline__ void
cuQuickTwoSum
	(double a, double b, double &s, double &e)
{
	s = __dadd_rn (a, b);
	e = __dsub_rn (b, __dsub_rn (s, a));
}

__device__ __forceinline__ void
cuSplit
	(double a, double &h, double &l) 
{
	double t;
	t = __dmul_rn (SPLITTER, a);
	h = __dsub_rn (t, __dsub_rn (t, a));
	l = __dsub_rn (a, h);
}

__device__ __forceinline__ void
cuTwoProd
	(double a, double b, double &p, double &e)
{
	double s1, s2, t1, t2;
	p = __dmul_rn (a, b);
	cuSplit(a, s1, s2);
	cuSplit(b, t1, t2);
	e = __dadd_rn (( __dsub_rn ( __dmul_rn (s1, t1), p)
	  + __dmul_rn (s1, t2) + __dmul_rn (s2, t1)), __dmul_rn (s2, t2));
}

__device__ __forceinline__ void
cuTwoProdFMA
	(double a, double b, double &p, double &e)
{
	p = __dmul_rn (a, b);
	e = __fma_rn (a, b, -p);
}


/* ***********************************************
   DD-Arithmetic
   *********************************************** */   

// ADD
__device__ __forceinline__ cuddreal
cuWaddV1
	(const cuddreal a, const cuddreal b)
{
	cuddreal t, s, c;
	cuTwoSum (a.x, b.x, t.x, t.y);
	cuTwoSum (a.y, b.y, s.x, s.y);
	t.y = __dadd_rn(t.y, s.x);
	cuQuickTwoSum (t.x, t.y, t.x, t.y);
	t.y = __dadd_rn (t.y, s.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWaddV2
	(const cuddreal a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoSum (a.x, b.x, t.x, t.y);
	t.y = __dadd_rn (t.y, __dadd_rn (a.y, b.y));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWaddV2
	(const double a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoSum (a, b.x, t.x, t.y);
	t.y = __dadd_rn (t.y, b.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWaddV2
	(const cuddreal a, const double b)
{
	cuddreal t;
	cuddreal c;
	cuTwoSum (a.x, b, t.x, t.y);
	t.y = __dadd_rn (t.y, a.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWaddV2
	(const double a, const double b)
{
	cuddreal c;
	cuTwoSum (a, b, c.x, c.y);
	return c;
}

// SUB
__device__ __forceinline__ cuddreal
cuWsubV1
	(const cuddreal a, const cuddreal b)
{
	cuddreal t, s, c;
	cuTwoSum (a.x, -b.x, t.x, t.y);
	cuTwoSum (a.y, -b.y, s.x, s.y);
	t.y = __dadd_rn(t.y, s.x);
	cuQuickTwoSum (t.x, t.y, t.x, t.y);
	t.y = __dadd_rn (t.y, s.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWsubV2
	(const cuddreal a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoSum (a.x, -b.x, t.x, t.y);
	t.y = __dadd_rn (t.y, __dsub_rn (a.y, b.y));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// MUL
__device__ __forceinline__ cuddreal
cuWmulV1
	(const cuddreal a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a.x, b.x, t.x, t.y);
	t.y = __dadd_rn (t.y, __dadd_rn (__dmul_rn(a.x, b.y), __dmul_rn (a.y, b.x)));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV1
	(const double a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a, b.x, t.x, t.y);
	t.y = __dadd_rn (t.y, __dmul_rn (a, b.y));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV1
	(const cuddreal a, const double b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a.x, b, t.x, t.y);
	t.y = __dadd_rn (t.y, __dmul_rn (a.y, b));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV1
	(const double a, const double b)
{
	cuddreal c;
	cuTwoProdFMA (a, b, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV2
	(const cuddreal a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a.x, b.x, t.x, t.y);
	t.y = __fma_rn(a.x, b.y, __fma_rn (a.y, b.x, t.y));
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV2
	(const double a, const cuddreal b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a, b.x, t.x, t.y);
	t.y = __fma_rn (a, b.y, t.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV2
	(const cuddreal a, const double b)
{
	cuddreal t;
	cuddreal c;
	cuTwoProdFMA (a.x, b, t.x, t.y);
	t.y = __fma_rn (a.y, b, t.y);
	cuQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

__device__ __forceinline__ cuddreal
cuWmulV3
	(const cuddreal a, const cuddreal b)
{
	double m1, t, e;
	cuddreal c;
	m1 = __dmul_rn(a.x, b.y);
	t = __fma_rn(a.y, b.x, m1);
	c.x = __fma_rn(a.x, b.x, t);
	e = __fma_rn(a.x, b.x, -c.x);
	c.y = __dadd_rn(e, t);
	return c;
}

#endif
