/* ***********************************************
 
	CSRMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_csrmv_n_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_CSRMV (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const size_t nnz,
	const TYPE0 *alpha,
	const TYPE0 *csrValA,
	const int32_t *csrRowPtrA,
	const int32_t *csrColIndA,
	const TYPE0 *x,
	const TYPE0 *beta,
	TYPE0 *y
) {
	void (*kernel) (const SIZE_T, const TYPE0, const TYPE0*, const int32_t*, const int32_t*, const TYPE0*, const TYPE0, TYPE0*);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	if (*alpha == ZERO0) {
		MUBLAS_SCAL (mh, m, beta, y, 1);
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	uint32_t kernel_number = 999;
	mh->m = m;
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	float nnzrow = (float)nnz/(float)m;
	mh->N_T = max(1, min(32, (size_t)pow(2.,ceil(log2(nnzrow)))));
	kernel_number = 100 + mh->N_T;
	// debug
	if (mh->kerneldebug != 0) {
		kernel_number = mh->kerneldebug;
		switch (kernel_number) {
			case 132: mh->N_T = 32; break;
			case 116: mh->N_T = 16; break;
			case 108: mh->N_T = 8; break;
			case 104: mh->N_T = 4; break;
			case 102: mh->N_T = 2; break;
			case 101: mh->N_T = 1; break;
		}
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		// cuSPARSE
		printf ("mublas error: not implemented");
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Launch Configuration
	if (trans == 'N' || trans == 'n') {
		enum cudaFuncCache func_cache;
		mh->BlkDim = 1;
		mh->func_getShmPerBlk = &getShmPerBlk_CSRMV_N;
		mh->func_getBlkPerGrd = &getBlkPerGrd_CSRMV_N;
		func_cache = cudaFuncCachePreferL1;
		#include "mublas_csrmv_n.incl"
	}
	switch (kernel_number) {
		case 132:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N32;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N32);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N32, func_cache);
			break;
		case 116:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N16;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N16);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N16, func_cache);
			break;
		case 108:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N8;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N8);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N8, func_cache);
			break;
		case 104:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N4;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N4);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N4, func_cache);
			break;
		case 102:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N2;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N2);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N2, func_cache);
			break;
		case 101:
			kernel = &MUBLAS_CSRMV_N_DEVICE_N1;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_CSRMV_N_DEVICE_N1);
			cudaFuncSetCacheConfig (MUBLAS_CSRMV_N_DEVICE_N1, func_cache);
			break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 6. Thread-Block Size Adjustment and Kernel Launch
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	if (mh->cc >= 300 || mh->N_T == 1) mh->ShmPerBlk = 0;
	if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, *alpha, csrValA, csrRowPtrA, csrColIndA, x, *beta, y);
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}
