/* ***********************************************
 
	GEAM-NN for CUDA 

   *********************************************** */   

#include "mublas_geam_nn.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEAM_NN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_B_GEAM_NN
#define RESTRICT_B __restrict__
#define LOAD1_B LOAD1_NC
#define LOAD2_B LOAD2_NC
#define LOAD4_B LOAD4_NC
#else
#define RESTRICT_B
#define LOAD1_B LOAD1___
#define LOAD2_B LOAD2___
#define LOAD4_B LOAD4___
#endif
#define STORE1_C STORE1___
#define STORE2_C STORE2___
#define STORE4_C STORE4___

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEAM_NN ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	mh->Bx = ceilingEx ((float)mh->m / (float)(mh->N_T * Bm), 1);
	mh->By = ceilingEx ((float)mh->n / (float)(Bn), 1);
}

__host__ __inline__ void getShmPerBlk_GEAM_NN ( mublasHandle_t* mh) {
	mh->ShmPerBlk = 0;
}

__host__ __inline__ size_t getTxSuf_GEAM_NN ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)mh->N_T / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	kernel=s1, trans=NN
   *********************************************** */   
__global__ void MUBLAS_GEAM_NN_DEVICE_s1 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 beta,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC1
	SIZE_T ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	
	ix = iBx * nTx + iTx;
	iy = iBy * nTy + iTy;

	TYPE1 Ar, Br, Cr;

	if (ix < m && iy < n) {
		LOAD1_A (Ar, &Ag[iy * lda + ix]);
		LOAD1_B (Br, &Bg[iy * ldb + ix]);
		Cr = MAD1 (alpha, Ar, MUL1 (beta, Br));
		STORE1_C (&Cg[iy * ldc + ix], Cr);
	}
#endif
}

/* ***********************************************
	kernel=v2, trans=NN
   *********************************************** */   
__global__ void MUBLAS_GEAM_NN_DEVICE_v2 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 beta,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	SIZE_T ix, iy;
	const SIZE_T N_T = 2;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	
	ix = (iBx * nTx + iTx) * N_T;
	iy = iBy * nTy + iTy;

	TYPE2 Ar2, Br2, Cr2;

	if (ix < m && iy < n) {
		LOAD2_A (Ar2, &Ag[iy * lda + ix]);
		LOAD2_B (Br2, &Bg[iy * ldb + ix]);
		Cr2 = MAD122 (alpha, Ar2, MUL12 (beta, Br2));
		STORE2_C (&Cg[iy * ldc + ix], Cr2);
	}
#endif
}

/* ***********************************************
	kernel=v4, trans=NN
   *********************************************** */   
__global__ void MUBLAS_GEAM_NN_DEVICE_v4 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 beta,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC4
	SIZE_T ix, iy;
	const SIZE_T N_T = 4;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	
	ix = (iBx * nTx + iTx) * N_T;
	iy = iBy * nTy + iTy;

	TYPE4 Ar4, Br4, Cr4;

	if (ix < m && iy < n) {
		LOAD4_A (Ar4, &Ag[iy * lda + ix]);
		LOAD4_B (Br4, &Bg[iy * ldb + ix]);
		Cr4 = MAD144 (alpha, Ar4, MUL14 (beta, Br4));
		STORE4_C (&Cg[iy * ldc + ix], Cr4);
	}
#endif
}


