/* ***********************************************
 
	GEMM for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemm_nn_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMM (
	mublasHandle_t* mh,	
	const char transA,
	const char transB,
	const size_t m,
	const size_t n,
	const size_t k,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *b,
	const size_t ldb,
	const TYPE0 *beta,
	TYPE0 *c,
	const size_t ldc
) {
	void (*kernel) (const SIZE_T, const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const SIZE_T, const TYPE0, TYPE0*, const SIZE_T);
	size_t kernel_number = 0;
	size_t use_vector = 0;
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) && (lda%mh->N_T == 0) &&
		(((long)b%(128/sizeof(TYPE0)))%16 == 0) && (ldb%mh->N_T == 0) &&
		(((long)c%(128/sizeof(TYPE0)))%16 == 0) && (ldc%mh->N_T == 0)
		) {
		use_vector = 1;
	} else {
		use_vector = 0;
	}
	mh->m = m;
	mh->n = n;
	mh->k = k;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Select
	kernel_number = 999; // default=CUBLAS;
	if (transB == 'N' || transB == 'n') {
		if (transA == 'N' || transA == 'n') {
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN;
			mh->func_getTxMax     = &getTxMax_GEMM_NN;
			mh->func_cache = cudaFuncCachePreferShared;
			#include "mublas_gemm_nn.incl"
			if (mh->N_T == 4 && use_vector) {
				kernel_number = 140;
			} else if (mh->N_T == 2 && use_vector) {
				kernel_number = 120;
			} else {
				kernel_number = 110;
			}
		}
	}
	// debug: 
	kernel_number = 110;
	mh->N_T = 1;
	if (mh->kerneldebug != 0)
		kernel_number = mh->kerneldebug;
	// ---------------------------------------------
		
	// ---------------------------------------------
	// 4. Unsupported operations (CUBLAS)
	if (kernel_number == 0) {
		#if defined (CUBLAS_GEMM)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		cublasOperation_t tranBcublas = CUBLAS_OP_N;
		if (transA == 'T' || transA == 't') tranAcublas = CUBLAS_OP_T;
		if (transA == 'C' || transA == 'c') tranAcublas = CUBLAS_OP_C;
		if (transB == 'T' || transB == 't') tranBcublas = CUBLAS_OP_T;
		if (transB == 'C' || transB == 'c') tranBcublas = CUBLAS_OP_C;
		CUBLAS_GEMM (ch, tranAcublas, tranBcublas, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Kernel Launch Configuration
	switch (kernel_number) {
		case 140:
			break;
		case 120:
			break;
		case 110:
			kernel = &MUBLAS_GEMM_NN_DEVICE_s1;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMM_NN_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMM_NN_DEVICE_s1, mh->func_cache);
			break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 6. Thread-Block Size Adjust and Kernel Launch
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	#ifdef USE_TEXTURE
    tex_mat_a.normalized = false;
	tex_mat_a.filterMode = cudaFilterModePoint;
	tex_mat_a.addressMode[0] = cudaAddressModeClamp;
	tex_mat_a.channelDesc = cudaCreateChannelDesc<int2>();
	tex_mat_b.normalized = false;
	tex_mat_b.filterMode = cudaFilterModePoint;
	tex_mat_b.addressMode[0] = cudaAddressModeClamp;
	tex_mat_b.channelDesc = cudaCreateChannelDesc<int2>();
	cudaBindTexture(0, tex_mat_a, a, sizeof(TYPE0) * lda * k);
	cudaBindTexture(0, tex_mat_b, b, sizeof(TYPE0) * ldb * n);
	#endif
	if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, k, *alpha, a, lda, b, ldb, *beta, c, ldc);
	#ifdef USE_TEXTURE
	cudaUnbindTexture (tex_mat_a);
	cudaUnbindTexture (tex_mat_b);
	#endif
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}
