/* ***********************************************
 
	GEMM-NN for CUDA 

   *********************************************** */   

#include "mublas_gemm_nn.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMM_NN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_B_GEMM_NN
#define RESTRICT_B __restrict__
#define LOAD1_B LOAD1_NC
#define LOAD2_B LOAD2_NC
#define LOAD4_B LOAD4_NC
#else
#define RESTRICT_B
#define LOAD1_B LOAD1___
#define LOAD2_B LOAD2___
#define LOAD4_B LOAD4___
#endif
#define LOAD1_C LOAD1___
#define STORE1_C STORE1___

#define NBM 1
#define NBN 1

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMM_NN ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * NBM;
	const SIZE_T Bn = mh->Ty * NBN;
	mh->Bx = ceilingEx ((float)mh->m / (float)(mh->N_T * Bm), 1);
	mh->By = ceilingEx ((float)mh->n / (float)(mh->N_T * Bn), 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * NBM;
	const SIZE_T Bn = mh->Ty * NBN;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_T * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_GEMM_NN ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)mh->N_T / (float)NBM / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	kernel=s1, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMM_NN_DEVICE_s1 (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC1
	SIZE_T jk, ik, ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	const SIZE_T Bm = nTx * NBM;
	const SIZE_T Bn = nTy * NBN;
	const SIZE_T Bk = nTx;
	const SIZE_T TsOffset = Bk*Bm;
	extern __shared__ TYPE1 Ts[];
	TYPE1 Tr[NBN][NBM], Cr;

	#pragma unroll
	for (iy = 0; iy < NBN; iy++) 
		#pragma unroll
		for (ix = 0; ix < NBM; ix++) 
			Tr[iy][ix] = ZERO1;

	for (jk = 0; jk < k; jk += Bk) {
		// Load sub-mat A from global to shared
		for (iy = 0; iy < Bk/nTy; iy++) {
			#pragma unroll
			for (ix = 0; ix < NBM; ix++) {
				LOAD1_A (
					Ts[(nTy*iy+iTy)*Bm+(nTx*ix+iTx)],
					&Ag[(jk+nTy*iy+iTy) * lda + (Bm*iBx+nTx*ix+iTx)]);
			}
		}
		// Load sub-mat B from global to shared
		#pragma unroll
		for (iy = 0; iy < NBN; iy++) {
			LOAD1_B (
				Ts[TsOffset+(NBN*iTy+iy)*Bk+(iTx)],
				&Bg[(Bn*iBy+NBN*iTy+iy) * ldb + (jk+iTx)]);
		}
		__syncthreads ();

		// sub-mat mul: C=A*B
		#pragma unroll
		for (ik = 0; ik < Bk; ik++) 
			#pragma unroll
			for (iy = 0; iy < NBN; iy++) 
				#pragma unroll
				for (ix = 0; ix < NBM; ix++) 
					Tr[iy][ix] = MAD1 (Ts[(ik)*Bm+(nTx*ix+iTx)], Ts[TsOffset+(NBN*iTy+iy)*Bk+(ik)], Tr[iy][ix]); 
		__syncthreads ();
	}

	// C
	#pragma unroll
	for (iy = 0; iy < NBN; iy++) {
		#pragma unroll
		for (ix = 0; ix < NBM; ix++) {
			LOAD1_C (
				Cr,
				&Cg[(Bn*iBy+NBN*iTy+iy) * ldc + (Bm*iBx+nTx*ix+iTx)]);
			Cr = MAD1 (alpha, Tr[iy][ix], MUL1 (beta, Cr));
			STORE1_C (
				&Cg[(Bn*iBy+NBN*iTy+iy) * ldc + (Bm*iBx+nTx*ix+iTx)],
				Cr);
		}
	}
#endif
}

/* ***********************************************
	NN-kernel (v4)
   *********************************************** */   
__global__ void MUBLAS_GEMM_NN_DEVICE_v4 (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
}

/* ***********************************************
	(simple) kernel=s1, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMM_NN_DEVICE_s1_ (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC1
	SIZE_T jk, ik, iTh, ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	const SIZE_T Bm = nTx * NBM;
	const SIZE_T Bn = nTy * NBN;
	const SIZE_T Bk = nTx;
	const SIZE_T TsOffset = Bk*Bm;
	extern __shared__ TYPE1 Ts[];
	TYPE1 Tr[NBN][NBM];

	#pragma unroll
	for (iy = 0; iy < NBN; iy++) 
		#pragma unroll
		for (ix = 0; ix < NBM; ix++) 
			Tr[iy][ix] = ZERO1;

	for (jk = 0; jk < k; jk += Bk) {
		for (iy = 0; iy < Bk/nTy; iy++) 
			#pragma unroll
			for (ix = 0; ix < NBM; ix++) 
				Ts[(nTy*iy+iTy)*Bm+(nTx*ix+iTx)] = Ag[(jk+nTy*iy+iTy) * lda + (Bm*iBx+nTx*ix+iTx)];
		#pragma unroll
		for (iy = 0; iy < NBN; iy++) 
			Ts[TsOffset+(NBN*iTy+iy)*Bk+(iTx)] = Bg[(Bn*iBy+NBN*iTy+iy) * ldb + (jk+iTx)];
		__syncthreads ();

		#pragma unroll
		for (ik = 0; ik < Bk; ik++) 
			#pragma unroll
			for (iy = 0; iy < NBN; iy++) 
				#pragma unroll
				for (ix = 0; ix < NBM; ix++) 
					Tr[iy][ix] = MAD1 (Ts[(ik)*Bm+(nTx*ix+iTx)], Ts[TsOffset+(NBN*iTy+iy)*Bk+(ik)], Tr[iy][ix]); 

		__syncthreads ();
	}

	#pragma unroll
	for (iy = 0; iy < NBN; iy++) {
		#pragma unroll
		for (ix = 0; ix < NBM; ix++) {
			iTh = (Bn*iBy+NBN*iTy+iy) * ldc + (Bm*iBx+nTx*ix+iTx);
			Cg[iTh] = MAD1 (alpha, Tr[iy][ix], MUL1 (beta, Cg[iTh]));
		}
	}
#endif
}


