/* ***********************************************
 
	GEMV-T for CUDA 

   *********************************************** */   

#include "mublas_gemv_t.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMV_T
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_X_GEMV_T
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif
#define LOAD1_Y LOAD1___
#define LOAD2_Y LOAD2___
#define LOAD4_Y LOAD4___
#define STORE1_Y STORE1___
#define STORE2_Y STORE2___
#define STORE4_Y STORE4___

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMV_T ( mublasHandle_t* mh) {
	mh->Bx = ceilingEx ((float)mh->n / (float)(mh->Tx / WrpSize), 1);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_GEMV_T ( mublasHandle_t* mh) {
	if (mh->cc >= 300)
		mh->ShmPerBlk = ceilingEx ((float)(mh->Tx * mh->N_one), mh->ShmUnit);
	else
		mh->ShmPerBlk = ceilingEx ((float)(mh->Tx * mh->N_one + WrpSize * mh->N_one * mh->N_T), mh->ShmUnit);
}

__host__ __inline__ size_t getTxSuf_GEMV_T ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->n * WrpSize / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	kernel=s1, trans=T
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_s1 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC1
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	const SIZE_T tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	const SIZE_T tid = blockDim.x * blockIdx.x + threadIdx.x;
	const SIZE_T rowid = tid / WrpSize;
	const SIZE_T lane  = tid % WrpSize;
	SIZE_T i;
	__shared__ TYPE1 Xs1[WrpSize];
	TYPE1 Ar1, Xr1, Yr1, Tr1;

	Tr1 = ZERO1;

	for (i = lane; i < m; i += WrpSize) {
		LOAD1_A (Ar1, &Ag[min(n-1, rowid) * lda + i]);
		LOAD1_X (Xr1, &Xg[i]);
		Xs1[lane] = Xr1;
		Tr1 = MAD1 (Ar1, Xs1[lane], Tr1);
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n) {
			LOAD1_Y (Yr1, &Yg[rowid]);
			Yr1 = MAD1 (alpha, Tr1, MUL1 (beta, Yr1));
			STORE1_Y (&Yg[rowid], Yr1);
		}
	}
#endif
}

/* ***********************************************
	kernel=v2, trans=T
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_v2 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC2
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	const SIZE_T tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	const SIZE_T N_T = 2;
	const SIZE_T tid = blockDim.x * blockIdx.x + threadIdx.x;
	const SIZE_T rowid = (tid / WrpSize);
	const SIZE_T lane  = tid % WrpSize;
	SIZE_T i, j, k;
	__shared__ TYPE2 Xs2[WrpSize];
	TYPE2 Ar2, Xr2, Tr2;
	TYPE1 Ar1, Xr1, Yr1, Tr1;

	Tr2 = ZERO2;
	Tr1 = ZERO1;

	for (i = 0; i < m / (WrpSize*N_T); i++) {
		j = i*WrpSize*N_T + lane*N_T;
		LOAD2_A (Ar2, &Ag[min(n-1, rowid) * lda + j]);
		LOAD2_X (Xr2, &Xg[j]);
		Xs2[lane] = Xr2;
		Tr2 = MAD2 (Ar2, Xs2[lane], Tr2);
	}
#if defined PREC_C || defined PREC_B || defined PREC_DB
	Tr1 = ADD1 ((TYPE1&)Tr2.x, (TYPE1&)Tr2.z);
#else
	Tr1 = ADD1 (Tr2.x, Tr2.y);
#endif

	if (m % (WrpSize*N_T) != 0) { 
		for (k = i*WrpSize*N_T + lane; k < m; k += WrpSize) {
			LOAD1_A (Ar1, &Ag[min(n-1, rowid) * lda + k]);
			LOAD1_X (Xr1, &Xg[k]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n) {
			LOAD1_Y (Yr1, &Yg[rowid]);
			Yr1 = MAD1 (alpha, Tr1, MUL1 (beta, Yr1));
			STORE1_Y (&Yg[rowid], Yr1);
		}
	}
#endif
}

/* ***********************************************
	kernel=v4, trans=T
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_v4 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC4
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	const SIZE_T tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	const SIZE_T N_T = 4;
	const SIZE_T tid = blockDim.x * blockIdx.x + threadIdx.x;
	const SIZE_T rowid = (tid / WrpSize);
	const SIZE_T lane  = tid % WrpSize;
	SIZE_T i, j, k;
	__shared__ TYPE4 Xs4[WrpSize];
	TYPE4 Ar4, Xr4, Tr4;
	TYPE1 Ar1, Xr1, Yr1, Tr1;

	Tr4 = ZERO4;
	Tr1 = ZERO1;

	for (i = 0; i < m / (WrpSize*N_T); i++) {
		j = i*WrpSize*N_T + lane*N_T;
		LOAD4_A (Ar4, &Ag[min(n-1, rowid) * lda + j]);
		LOAD4_X (Xr4, &Xg[j]);
		Xs4[lane] = Xr4;
		Tr4 = MAD4 (Ar4, Xs4[lane], Tr4);
	}
	Tr1 = ADD1 (ADD1 (Tr4.x, Tr4.y), ADD1 (Tr4.z, Tr4.w));

	if (m % (WrpSize*N_T) != 0) { 
		for (k = i*WrpSize*N_T + lane; k < m; k += WrpSize) {
			LOAD1_A (Ar1, &Ag[min(n-1, rowid) * lda + k]);
			LOAD1_X (Xr1, &Xg[k]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n) {
			LOAD1_Y (Yr1, &Yg[rowid]);
			Yr1 = MAD1 (alpha, Tr1, MUL1 (beta, Yr1));
			STORE1_Y (&Yg[rowid], Yr1);
		}
	}
#endif
}

#undef RESTRICT_A
#undef RESTRICT_X
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef LOAD1_Y
#undef LOAD2_Y
#undef LOAD4_Y
#undef STORE1_Y
#undef STORE2_Y
#undef STORE4_Y
