/* ***********************************************
 
	TRMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_trmv_lnn_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_TRMV (
	mublasHandle_t* mh,	
	const char uplo,
	const char trans,
	const char diag,
	const size_t n,
	const TYPE0 *a,
	const size_t lda,
	TYPE0 *x,
	const size_t incx
) {
	uint32_t kernel_number = 0;
	void (*kernel) (const SIZE_T, const TYPE0*, const SIZE_T, const TYPE0*, TYPE0*, TYPE0*);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (n == 0) return 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration
	kernel_number = 999;
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)a%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)x%(128/sizeof(TYPE0)))%16 != 0) ||
		(lda%mh->N_T != 0)) mh->N_T = 1;
	if (incx == 1) {
		if (uplo == 'L' || uplo == 'l') {
			if (trans == 'N' || trans == 'n') {
				if (diag == 'N' || diag == 'n') 
					kernel_number = 100 + mh->N_T * 10;
			}
		}
	}
	// debug
	if (mh->kerneldebug != 0) {
		kernel_number = mh->kerneldebug;
		switch (kernel_number) {
			case 110: mh->N_T = 1; break;
			case 120: mh->N_T = 2; break;
			case 140: mh->N_T = 4; break;
		}
	}
	// ---------------------------------------------
	
	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_TRMV)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasFillMode_t uploAcublas = CUBLAS_FILL_MODE_UPPER;
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		cublasDiagType_t diagAcublas = CUBLAS_DIAG_NON_UNIT;
		if (uplo == 'L' || uplo == 'l') uploAcublas = CUBLAS_FILL_MODE_LOWER;
		if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
		if (diag == 'U' || trans == 'u') diagAcublas = CUBLAS_DIAG_UNIT;
		CUBLAS_TRMV (ch, uploAcublas, tranAcublas, diagAcublas, n, a, lda, x, incx);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Launch Configuration
	if (uplo == 'L' || uplo == 'l') {
		if (trans == 'N' || trans == 'n') {
			if (diag == 'N' || diag == 'n') {
				mh->BlkDim = 2;
				mh->func_getShmPerBlk = &getShmPerBlk_TRMV_LNN;
				mh->func_getBlkPerGrd = &getBlkPerGrd_TRMV_LNN;
				mh->func_cache = cudaFuncCachePreferShared;
				#include "mublas_trmv_lnn.incl"
			}
		}
	}
	switch (kernel_number) {
		case 140:
			kernel = &MUBLAS_TRMV_LNN_DEVICE_v4;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_v4, mh->func_cache);
			break;
		case 120:
			kernel = &MUBLAS_TRMV_LNN_DEVICE_v2;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_v2, mh->func_cache);
			break;
		case 110:
			kernel = &MUBLAS_TRMV_LNN_DEVICE_s1;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_TRMV_LNN_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_TRMV_LNN_DEVICE_s1, mh->func_cache);
			break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Thread-Block Size Adjustment and Kernel Launch
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	dim3 threads_l1 = dim3 (64);
	dim3 grid_l1 = dim3 ((size_t) ceil ((float)mh->n / 64.));
	if (mh->autotuning!=1) {
		kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (n, a, lda, x, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2);
		MUBLAS_TRMV_LNN_DEVICE_final <<< grid_l1, threads_l1, 0, mh->streamId >>> (n, (TYPE0*)mh->work_addr, (TYPE0*)mh->work_addr2, x);
	}
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

